(* Arguably, "Not" may merit its own type since it is unary *)
type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq
          | Mod | Not | And | Or | Shl | Shr | Concat | Deref

type direction = Up | Down
type interval = Whole | Half

type primitive =   Bool | Int | Note | List of primitive
                 | Interval | Void | Safe | Illegal

type expr =
    Literal of int
  | BoolLiteral of bool
  | NoteLiteral of int
  | ListLiteral of expr list
  | StringLiteral of string
  | EmptyList of primitive
  | IntervalLiteral of interval
  | Id of string
  | Binop of expr * op * expr
  | Unaryop of op * expr
  | Assign of string * expr
  | Call of string * expr list
  | StepExpr of direction * interval
  | Noexpr

type stmt =
    Block of stmt list
  | Declaration of string * expr * bool (* identifier, value, in foreach? *)
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of string * expr * expr * expr * stmt (* id, initial, limit, step, body *)
  | ForEach of string * expr * stmt (* refactor as for? *)
  | While of expr * stmt

type func_decl = {
    rettype : primitive;
    fname : string;
    formals : (primitive * string) list;
    body : stmt list;
  }

type variable_decl = {
  vname : string;
  value : expr;
}

type program = variable_decl list * func_decl list

let rec string_of_type t = match t with
    Bool -> "bool"
  | Int -> "int"
  | Note -> "note"
  | List(t) -> "list<" ^ string_of_type t ^ ">"
  | Interval -> "interval"
  | Void -> "void"
  | Safe -> "safe"
  | Illegal -> "illegal"

let notestring_of_int i = (match i with
        0 -> "A0"
      | 1 -> "A#0"
      | 2 -> "B0"
      | _ -> (match ((i-3) mod 12) with
          0 -> "C"
        | 1 -> "C#"
        | 2 -> "D"
        | 3 -> "D#"
        | 4 -> "E"
        | 5 -> "F"
        | 6 -> "F#"
        | 7 -> "G"
        | 8 -> "G#"
        | 9 -> "A"
        | 10 -> "A#"
        | 11 -> "B"
        | _ -> raise (Failure("oob note"))) ^ (string_of_int (((i-3)/12)+1))
      )

let index_in_key note = match note with
    'C' -> 0
  | 'D' -> 2
  | 'E' -> 4
  | 'F' -> 5
  | 'G' -> 7
  | 'A' -> 9
  | 'B' -> 11
  | _ -> raise (Failure("invalid note")) ;;

let accidental_modifier = function
    'b' -> (-1)
  | '#' -> 1
  | _ -> raise (Failure("invalid accidental"))

let noteint_of_str s = match s with
    "A0" -> 0
  | "A#0" -> 1
  | "Bb0" -> 1
  | "B0" -> 2
  | "B#0" -> 3
  | _ -> 3 + (match String.length s with
      1 -> 36 + (index_in_key (String.get s 0))
    | 2 -> (match String.get s 1 with
          'b' -> 36 + (index_in_key (String.get s 0)) - 1
        | '#' -> 36 + (index_in_key (String.get s 0)) + 1
        | _ -> (index_in_key (String.get s 0) +
               (12 * ((int_of_char(String.get s 1) - 1) - int_of_char('0')))))
    | 3 -> (index_in_key (String.get s 0) +
            accidental_modifier (String.get s 1) +
            (12 * ((int_of_char(String.get s 2) - 1) - int_of_char('0'))))
    | _ -> raise (Failure("invalid syntax"))) ;;


let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | BoolLiteral(false) -> "false"
  | BoolLiteral(true) -> "true"
  | NoteLiteral(n) -> notestring_of_int n
  | EmptyList(t) -> "list<" ^ string_of_type t ^ ">"
  | ListLiteral(l) ->
          "[" ^ (String.concat ";" (List.map string_of_expr l)) ^ "]"
  | StringLiteral(s) -> "\"" ^ String.escaped s ^ "\""
  | IntervalLiteral(s) ->
                  (match s with Whole -> "whole" | Half -> "half")
  | Id(s) -> s
  | Binop(e1, Deref, e2) -> "(" ^ string_of_expr e1 ^ "[" ^
      string_of_expr e2 ^ "])"
  | Binop(e1, o, e2) -> "(" ^
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">="
      | Shr -> ">>" | Shl -> "<<" | Or -> "||" | And -> "&&"
      | Not -> "!" | Mod -> "%" | Concat -> "::"
      | _ -> raise(Failure("illegal binop"))) ^ " " ^
      string_of_expr e2 ^ ")"
  | Unaryop(o, e) -> "(" ^
      (match o with
        Not -> "!"
      | _ -> raise (Failure("illegal unary operator")))
      ^ string_of_expr e ^ ")"
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | StepExpr(d, i) ->
                  (match d with Up -> "up" | Down -> "down") ^ " " ^
                  (match i with Whole -> "whole" | Half -> "half")
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Declaration(i, e, _) -> i ^ " := " ^ string_of_expr e ^ ";\n"
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(i, e1, e2, e3, s) -> "for (" ^ i ^ " := " ^ string_of_expr e1 ^
      " to " ^ string_of_expr e2 ^
      (match e3 with Noexpr -> "" | _ -> " : " ^ string_of_expr e3) ^ ") {\n" ^
      string_of_stmt s ^ "\n}\n"
  | ForEach(i, e, s) -> "for (" ^ i ^ " in " ^ string_of_expr e ^ ") {\n" ^
      string_of_stmt s ^ "\n}\n"
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s

let string_of_vdecl id = id.vname ^ " := " ^ string_of_expr id.value
                          ^ ";\n"

let string_of_formal f = string_of_type (fst f) ^ " " ^ snd f
let string_of_formals f = String.concat "," (List.map string_of_formal f)

let string_of_fdecl fdecl =
  string_of_type fdecl.rettype ^ " " ^ fdecl.fname ^ "(" ^
  string_of_formals (fdecl.formals) ^ ")\n{\n" ^
(*  String.concat "" (List.map string_of_vdecl fdecl.locals) ^ *)
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
