open Ast
type bstmt =
    Lit of int    (* Push a literal *)
  | Sng of string (* Push a string *)
  | Nte of int    (* Push a note *)
  | Bln of bool   (* Push a boolean *)
  | Ivl of Ast.interval (* Push an interval *)
  | Lst of (Ast.primitive * int * (int list))
  | Drp           (* Discard a value *)
  | Dup           (* Duplicate top of stack *)
  | Efr           (* Enter a foreach block *)
  | Swp           (* swap two top stack arguments *)
  | Bin of Ast.op (* Perform arithmetic on top of stack *)
  | Una of Ast.op (* Perform a unary operation *)
  | Dbg           (* dump stack *)
  | Nop           (* no-op *)
  | Lod of int    (* Fetch global variable *)
  | Str of int    (* Store global variable *)
  | Lfp of int    (* Load frame pointer relative *)
  | Sfp of int    (* Store frame pointer relative *)
  | Jsr of int    (* Call function by absolute address *)
  | Ent of int    (* Push FP, FP -> SP, SP += i *)
  | Rts of int    (* Restore FP, SP, consume formals, push result *)
  | Beq of int    (* Branch relative if top-of-stack is true *)
  | Bne of int    (* Branch relative if top-of-stack is false *)
  | Bra of int    (* Branch relative *)
  | Hlt           (* Terminate *)

type prog = {
    num_globals : int;   (* Number of global variables *)
    text : bstmt array; (* Code for all the functions *)
  }

let string_of_stmt = function
    Lit(i) -> "Lit " ^ string_of_int i
  | Sng(s) -> "Sng " ^ String.escaped s
  | Nte(n) -> "Nte " ^ notestring_of_int n
  | Bln(b) -> "Bln " ^ string_of_bool b
  | Lst(t, l, c) -> "Lst " ^ string_of_type t ^ "(" ^ string_of_int l ^ " els)"
  | Ivl(Half) -> "Ivl Half"
  | Ivl(Whole) -> "Ivl Whole"
  | Drp -> "Drp"
  | Bin(Ast.Add) -> "Add"
  | Bin(Ast.Sub) -> "Sub"
  | Bin(Ast.Mult) -> "Mul"
  | Bin(Ast.Div) -> "Div"
  | Bin(Ast.Equal) -> "Eql"
  | Bin(Ast.Neq) -> "Neq"
  | Bin(Ast.Less) -> "Lt"
  | Bin(Ast.Leq) -> "Leq"
  | Bin(Ast.Greater) -> "Gt"
  | Bin(Ast.Geq) -> "Geq"
  | Bin(Ast.Deref) -> "Drf"
  | Bin(Ast.Concat) -> "Con"
  | Bin(Ast.Shr) -> "Shr"
  | Bin(Ast.Shl) -> "Shl"
  | Bin(Ast.Or) -> "Or"
  | Bin(Ast.And) -> "And"
  | Bin(Ast.Mod) -> "Mod"
  | Una(Ast.Not) -> "Not"
  | Dbg -> "Dbg"
  | Nop -> "Nop"
  | Dup -> "Dup"
  | Efr -> "Efr"
  | Swp -> "Swp"
  | Lod(i) -> "Lod " ^ string_of_int i
  | Str(i) -> "Str " ^ string_of_int i
  | Lfp(i) -> "Lfp " ^ string_of_int i
  | Sfp(i) -> "Sfp " ^ string_of_int i
  | Jsr(i) -> "Jsr " ^ string_of_int i
  | Ent(i) -> "Ent " ^ string_of_int i
  | Rts(i) -> "Rts " ^ string_of_int i
  | Bne(i) -> "Bne " ^ string_of_int i
  | Beq(i) -> "Beq " ^ string_of_int i
  | Bra(i) -> "Bra " ^ string_of_int i
  | Hlt    -> "Hlt"
  | _ -> raise (Failure("unimplemented instruction"))

let string_of_prog p =
  string_of_int p.num_globals ^ " global variables\n" ^
  let funca = Array.mapi
      (fun i s -> string_of_int i ^ " " ^ string_of_stmt s) p.text
  in String.concat "\n" (Array.to_list funca)
