open Ast
open Bytecode

module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
    (* Convenience map from function name -> declaration *)
    fmap : func_decl StringMap.t;
    (* Map from function name -> (index, declaration) *)
    function_index : (int * func_decl) StringMap.t;
    (* "Address" for global variables *)
    global_index   : (int * primitive) StringMap.t;
    (* FP offset for args *)
    formal_index : (int * primitive) StringMap.t;
    (* FP offset for locals *)
    local_index    : (int * primitive) StringMap.t;
    declared_vars : bool array;
  }

(* Determines type based on function declaration / scopes *)
let rec get_type expr env = match expr with
    Literal(_) -> Int
  | BoolLiteral(_) -> Bool
  | NoteLiteral(_) -> Note
  | IntervalLiteral(_) -> Interval
  (* "Illegal" can still be passed to "Safe" *)
  | StringLiteral(_) -> Illegal
  | EmptyList(t) -> List(t)
  | ListLiteral(e) ->
      let t = (List.fold_left (fun t e -> (match (t, get_type e env) with
        (Illegal, x) -> x
      | (x, y) -> if x = y then x
                  else raise(Failure("conflict types in list literal"))))
      Illegal e) in
              (match t with
                Int -> List(Int) | Bool -> List(Bool) | Note -> List(Note)
              | _ -> raise(Failure("illegal list type")))
  | Unaryop(_, e) -> get_type e env
  | Binop(e1, op, e2) ->
      (let t1 = get_type e1 env and
           t2 = get_type e2 env in match (t1, op, t2) with
      | (x, op, y)
          when (x = y &&
               (op = Equal || op = Neq) &&
               (x = Int || x = Note || x = Bool)) -> Bool
      | (x, op, y)
          when (x = y &&
               (op = Leq || op = Geq || op = Less || op = Greater) &&
               (x = Int || x = Note)) -> Bool
      | (Bool, op, Bool) when (op = And || op = Or) -> Bool
      | (Note, op, Interval) when (op = Add || op = Sub) -> Note
      | (Note, op, Int) when (op = Shl || op = Shr) -> Note
      | (Int, op, Int)
          when (op = Add || op = Sub || op = Mult ||
                op = Div || op = Mod) -> Int
      | (x, Concat, List(y)) when (x = y) -> List(x)
      | (List(t), Deref, Int) -> t
      | _ -> raise(Failure("illegal types to binary operator")))
  | Call(f, formals) -> (match f with
        "reverse" -> get_type (List.nth formals 0) env
      | _ -> (StringMap.find f env.fmap).rettype)
  | Noexpr -> Void
  | Assign(_) -> Void
  | Id(s) ->
      (try snd(StringMap.find s env.local_index)
        with Not_found -> try snd(StringMap.find s env.formal_index)
        with Not_found -> try snd(StringMap.find s env.global_index)
        with Not_found -> raise(Failure("type undeclared variable " ^ s)))
  | _ -> Illegal

(* val enum : int -> 'a list -> (int * 'a) list *)
let rec enum stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(* Count # of local declarations in a function (to allocate stack) *)
let rec sum_decls body =
  let rec sum_decl sum stmt =
      match stmt with
        Declaration(_) -> sum + 1
      | Block(s) -> sum_decls s
      | For(i, e1, e2, e3, s) -> (sum + 1) + (sum_decl 0 s)
      | ForEach(i, e, s) -> sum + 2 + (sum_decl 0 s)
      | If(p, t, e) -> sum + (sum_decl 0 t) + (sum_decl 0 e)
      | While(e, s) -> sum + (sum_decl 0 s)
      | _ -> sum
    in List.fold_left sum_decl 0 body

(* Create list of local declarations (to determine FP offsets) *)
let decl_list body =
  let rec collect_decl l stmt = match stmt with
      Declaration(i, e, _) as d -> d::l
    | For(i, e1, e2, e3, s) -> collect_decl ((Declaration(i, e1, false))::l) s
    (* 
     * The "true" 3rd argument indicates this declaration is from a
     * ForEach, so when we do type checking, we know that the type is
     * actually a Foo, not a List<Foo>
     *)
    | ForEach(i, e, s) -> collect_decl ((Declaration(i, e, true))::l) s
    | Block(s) -> List.fold_left collect_decl l s
    | If(p, t, e) -> collect_decl (collect_decl l e) t
    | While(e, s) -> collect_decl l s
    | _ -> l in
    List.rev (List.fold_left collect_decl [] body)

(* Create a list of formals (to determine FP offsets) *)
let list_formals formals = List.fold_left (fun l d -> (snd d)::l) [] formals

let builtins = [
        { rettype = Int;
          fname = "len";
          formals = [(List(Safe), "arg")];
          body = []
        };
        { rettype = Int;
          fname = "octave";
          formals = [(Note, "n")];
          body = []
        };
        { rettype = Note;
          fname = "pitch";
          formals = [(Note, "n")];
          body = []
        };
        { rettype = Void;
          fname = "print";
          formals = [(Safe, "p")];
          body = [];
        };
        { rettype = Safe;
          fname = "reverse";
          formals = [(List(Safe), "list")];
          body = [];
        }]

(* returns (int * primitive) - address / type for global vdecls *)
let compute_global_index decls env funcs =
  let decls =
    List.map (fun (n, {vname=i; value=e}) ->
            (n, Declaration(i, e, false))) decls in
  let rec build_index decls env =
    match decls with
    | [(n, Declaration(i, e, _))] ->
        StringMap.add i (n, get_type e env) env.global_index
    | (n, Declaration(i, e, _))::tl ->
        let new_idx =
          StringMap.add i (n, get_type e env) env.global_index in
        build_index tl { env with global_index = new_idx }
    | [] -> env.global_index
    | _ -> raise(Failure("illegal argument to compute_index"))
  in build_index decls env

let check_declaration id env =
  let idx =
    try (fst(StringMap.find id env.local_index))
    with Not_found -> try (fst(StringMap.find id env.formal_index))
    with Not_found -> -1 in
  let is_global = try (ignore (StringMap.find id env.global_index); true)
                  with not_found -> false in

  if (idx > 0 && env.declared_vars.(idx - 1) = false) then
    raise(Failure("variable " ^ id ^ " not yet declared"))
  else if (idx = 0 && is_global = false) then
    raise(Failure("undeclared variable " ^ id))
  else ()

let declare id env =
  let idx = (fst(StringMap.find id env.local_index)) in
    if (env.declared_vars.(idx - 1) = false) then
      env.declared_vars.(idx - 1) <- true
    else
      raise(Failure("redeclaration of " ^ id))

(* does the same as compute_global_index but for local variables *)
let compute_local_index decls env funcs =
  let rec build_index decls env =
    match decls with
    | [(n, Declaration(i, e, false))] ->
        StringMap.add i (n, get_type e env) env.local_index
    | [(n, Declaration(i, e, true))] ->
        let t = (match get_type e env with
            List(Int) -> Int
          | List(Note) -> Note
          | List(Bool) -> Bool
          | _ -> raise(Failure("invalid type in foreach"))) in
        StringMap.add i (n, t) env.local_index
    | (n, Declaration(i, e, false))::tl ->
        let new_idx =
          StringMap.add i (n, get_type e env) env.local_index in
        build_index tl { env with local_index = new_idx }
    | (n, Declaration(i, e, true))::tl ->
        let t = (match get_type e env with
            List(Int) -> Int
          | List(Note) -> Note
          | List(Bool) -> Bool
          | _ -> raise(Failure("invalid type in foreach"))) in
        let new_idx =
          StringMap.add i (n, t) env.local_index in
        build_index tl { env with local_index = new_idx }
    | [] -> env.local_index
    | _ -> raise(Failure("illegal argument to compute_index"))
  in build_index decls env

(* and the same, but for formals *)
let compute_formal_index formals =
    List.fold_left
      (fun m (n, (t, i)) -> StringMap.add i (n, t) m) StringMap.empty formals

let len_arg_check t1 t2 = match (t1, t2) with
    (List(Safe), List(_)) -> true
  | (List(_), List(Safe)) -> true
  | _ -> false

let check_function_args actuals f env =
  let arg_types = List.map (fun formal -> fst formal) f.formals in
  List.map2
  (fun e t2 -> let t1 = get_type e env in
  if t1 = Safe || t2 = Safe || (t1 = t2) || (len_arg_check t1 t2) then true
  else
    raise(Failure("invalid arg type in call to " ^ f.fname)))
  actuals arg_types

let default_return fdecl = match fdecl.rettype with
    Bool -> [Bln false]
  | Int -> [Lit 0]
  | Void -> []
  | Note -> [Nte (noteint_of_str "C4")]
  | List(t) -> [Lst (t, 0, [])]
  | _ -> raise(Failure("illegal return type for " ^ fdecl.fname))

let check_assign id expr env =
  (* verify it was declared *)
  check_declaration id env;
  let t1 = get_type expr env and
      t2 = (try snd(StringMap.find id env.local_index)
      with Not_found -> try snd(StringMap.find id env.formal_index)
      with Not_found -> try snd(StringMap.find id env.global_index)
      with Not_found -> raise (Failure ("undeclared variable " ^ id))) in
  if (not(t1 = t2) || (t1 = Void || t1 == Illegal)) then
    raise(Failure("illegal type in assignment to " ^ id ^
          "(" ^ (string_of_type t2) ^ " = " ^ (string_of_type t1) ^ ")"))

(** Translate a program in AST form into a bytecode program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or function *)
let translate (globals, functions) =
  let fmap = List.fold_left
    (fun m f -> StringMap.add f.fname f m) StringMap.empty functions in
  let fmap = List.fold_left
    (fun m f -> StringMap.add f.fname f m) fmap builtins in
  (* pseudo-function to initialize global variables *)
  let start_fdecl =
          { rettype = Void;
            fname = "$start$";
            formals = [];
            body = List.map
            (function { vname=id; value=expr } -> Expr(Assign(id, expr)))
            globals
          } in
  let functions = start_fdecl::functions in
  (* Assign indexes to function names; built-ins are special *)

  let function_indexes =
    let bf = enum (-1) (-1) builtins and uf = enum 1 1 functions in
      List.fold_left
      (fun m (i, f) -> StringMap.add f.fname (i, f) m)
        StringMap.empty (bf @ uf) in
  (* Allocate "addresses" for each global variable and assign types *)
  let global_indexes =
    (* gi = (0, (Declaration("foo", Expr(Id("i"))) *)
    let gi = enum 1 0 globals in
    let stubenv =
            { fmap = fmap;
              function_index = function_indexes;
              global_index = StringMap.empty;
              formal_index = StringMap.empty;
              local_index = StringMap.empty;
              declared_vars = Array.make 0 false;
          } in
    compute_global_index gi stubenv fmap in

  let sm_length = (StringMap.fold (fun k v s -> s + 1) global_indexes 0) in
  let verify_globals =
    if (List.length globals != sm_length) then
      raise (Failure("global variable redeclaration")) in
      ignore verify_globals;

  (* Translate a function in AST form into a list of bytecode statements *)
  let translate env fdecl =
    (* Bookkeeping: FP offsets for locals and arguments *)
    let num_formals = List.length fdecl.formals
    and num_locals = sum_decls fdecl.body
    (* this must come first to allow initialization of locals with formals *)
    in let env = { env with
      formal_index = compute_formal_index (enum (-1) (-2) fdecl.formals);
    }
    in let env = { env with
      local_index =
        compute_local_index (enum 1 1 (decl_list fdecl.body)) env fmap;
      declared_vars = Array.make num_locals false } in
    let rec expr = function
	Literal i -> [Lit i]
      | StringLiteral s -> [Sng s]
      | NoteLiteral n -> [Nte n]
      | BoolLiteral b -> [Bln b]
      | IntervalLiteral i -> [Ivl i]
      | EmptyList(t) -> [Lst(t, 0, [])]
      | ListLiteral(el) as ll -> let
          values =
            (List.fold_left (fun bc e -> bc @ expr e) [] el) and
          ldecl = (match get_type ll env with
              List(t) -> [Lst(t, 0, [])]
            | _ -> raise(Failure("illegal list type"))) and
          cons =
              (List.fold_left (fun bc i -> Bin(Ast.Concat)::bc) [] el)
          in values @ ldecl @ cons
      | Id s ->
	  (try [Lfp (fst(StringMap.find s env.local_index))]
          with Not_found -> try [Lfp (fst(StringMap.find s env.formal_index))]
          with Not_found -> try [Lod (fst(StringMap.find s env.global_index))]
          with Not_found -> raise (Failure ("undeclared variable " ^ s)))
      | Unaryop (op, e) -> expr e @ [Una op]
      | Binop (e1, op, e2) as b ->
                      ignore (get_type b env); (* check type of expression *)
                      expr e1 @ expr e2 @ [Bin op]
      | Assign (s, e) -> check_assign s e env; expr e @
	  (try [Sfp (fst(StringMap.find s env.local_index))]
      with Not_found -> try [Sfp (fst(StringMap.find s env.formal_index))]
      with Not_found -> [Str (fst(StringMap.find s env.global_index))])
      | Call (fname, actuals) -> (try
        let f = StringMap.find fname env.function_index in
          ignore (check_function_args actuals (snd f) env);
          (List.concat (List.map expr (List.rev actuals))) @
          [Jsr (fst f) ]
        with Not_found -> raise (Failure ("undefined function " ^ fname)))
      | Noexpr -> []
      | _ -> raise(Failure("unimplemented expr translation"))

    in let rec stmt = function
	Block sl     ->  List.concat (List.map stmt sl)
      | Declaration(i, e, _) -> declare i env; expr (Assign(i, e))
      | Expr e       -> expr e @ [Drp]
      | Return e     ->
        if (fdecl.rettype = (get_type e env)) then
          expr e @ [Rts num_formals]
        else
          raise(Failure("illegal return type in " ^ fdecl.fname))
      | If (p, t, f) -> let t' = stmt t and f' = stmt f in
        if ((get_type p env) = Bool) then
          expr p @ [Bne(2 + List.length t')] @
          t' @ [Bra(1 + List.length f')] @ f'
        else
          raise(Failure("non-bool argument to if statement"))
      | For(i, e1, e2, e3, b) ->
          let d = Declaration(i, e1, false)
          and p = Binop(Id(i), Neq, e2)
          and s = (let t = (get_type (Id(i)) env) in
            match (t, e3) with
              (Int, Noexpr) -> Assign(i, Binop(Id(i), Add, Literal(1)))
            | (Note, Noexpr) -> Assign(i, Binop(Id(i), Add,
                                IntervalLiteral(Half)))
            | (Note, StepExpr(Up, ivl)) -> Assign(i, Binop(Id(i), Add,
                                IntervalLiteral(ivl)))
            | (Note, StepExpr(Down, ivl)) -> Assign(i, Binop(Id(i), Sub,
                                IntervalLiteral(ivl)))
            | (_, _) -> Assign(i, Binop(Id(i), Add, e3))) in
          stmt (Block([d; While(p, Block([b; Expr(s)]))]))
      | ForEach(i, e, b) -> declare i env;
        let t = get_type e env in (match t with
            List(_) -> ()
          | _ -> raise(Failure("cannot call foreach on non-list")));
          let b' = stmt b and e' = expr e and u = [Lit 1; Bin(Add)] in
        let prologue = [Efr; Lit (-1)]
        and epilogue = [Drp; Drp]
        and test = [Dup] @ e' @ [Jsr(-1); Bin(Equal)]
        and asn = [Dup] @ e' @ [Swp; Bin(Deref);
                   Sfp (fst(StringMap.find i env.local_index)); Drp] in
      prologue @
      [Bra(1+List.length b' + List.length asn)] @ asn @ b' @ u @ test @
      [Bne(-(List.length asn + List.length b' +
             List.length u + List.length test))] @
      epilogue
      | While (e, b) ->
	  let b' = stmt b and e' = expr e in
	  [Bra (1+ List.length b')] @ b' @ e' @
	  [Beq (-(List.length b' + List.length e'))]
    in [Ent num_locals] @      (* Entry: allocate space for locals *)
    stmt (Block fdecl.body) @  (* Body *)
    default_return fdecl @
    [Rts num_formals]
  in let env = { fmap = fmap;
     function_index = function_indexes;
		 global_index = global_indexes;
     formal_index = StringMap.empty;
     local_index = StringMap.empty;
     declared_vars = (Array.make 0 false)} in
  (* Code executed to start the program: Jsr $start$; Jsr main; halt *)
  let entry_function = try
    [Jsr (fst (StringMap.find "$start$" function_indexes));
     Jsr (fst (StringMap.find "main" function_indexes));
     Hlt]
  with Not_found -> raise (Failure ("no \"main\" function"))
  in
    
  (* Compile the functions *)
  let func_bodies = entry_function :: List.map (translate env) functions in

  (* Calculate function entry points by adding their lengths *)
  let (fun_offset_list, _) = List.fold_left
      (fun (l,i) f -> (i :: l, (i + List.length f))) ([],0) func_bodies in
  let func_offset = Array.of_list (List.rev fun_offset_list) in

  { num_globals = List.length globals;
    (* Concatenate the compiled functions and replace the function
       indexes in Jsr statements with PC values *)
    text = Array.of_list (List.map (function
	Jsr i when i > 0 -> Jsr func_offset.(i)
      | _ as s -> s) (List.concat func_bodies))
  }
