open Ast
open Bytecode

type stack_contents =
    StackInt of int
  | StackNte of int
  | StackSng of string
  | StackBln of bool
  | StackIvl of Ast.interval
  | StackPos of int
  | StackLst of primitive * int * (int list)
  | RetAddr of int

let string_of_stack item = match item with
    StackInt(i) -> "StackInt(" ^ string_of_int(i) ^ ")"
  | StackNte(i) -> "StackNte(" ^ string_of_int(i) ^ ")"
  | StackSng(s) -> "StackSng(" ^ s ^ ")"
  | StackBln(b) -> "StackBln(" ^ string_of_bool(b) ^ ")"
  | StackIvl(Half) -> "StackIvl(Half)"
  | StackIvl(Whole) -> "StackIvl(Whole)"
  | StackPos(i) -> "StackPos(" ^ string_of_int i ^ ")"
  | StackLst(t, l, v) -> "StackLst(" ^ (string_of_type t) ^ ", " ^
                          string_of_int(l) ^ " els)"
  | RetAddr(i) -> "RetAddr(" ^ string_of_int(i) ^ ")"

let rec print_stack sp pos len =
  if len > 0 then
    (print_endline ("sp(" ^ (string_of_int pos) ^ ") = " ^
    string_of_stack (sp.(pos-1))); print_stack sp (pos-1) (len-1))
  else ()

let ret_addr s = match s with
    RetAddr(i) -> i
  | _ -> raise (Failure("trying to compute ret address from non-address"))

let stack_pos s = match s with
    StackPos(i) -> i
  | _ -> raise (Failure("trying to compute stack address from non-address"))

(* Stack layout just after "Ent":

              <-- SP
   Local n
   ...
   Local 0
   Saved FP   <-- FP
   Saved PC
   Arg 0
   ...
   Arg n *)

let execute_prog prog =
  let stack = Array.make 1024 (StackInt(0))
  and globals = Array.make prog.num_globals (StackInt(0)) in

  let rec exec fp sp pc = match prog.text.(pc) with
    Lit i  -> stack.(sp) <- StackInt(i) ; exec fp (sp+1) (pc+1)
  | Sng s  -> stack.(sp) <- StackSng(s); exec fp (sp+1) (pc+1)
  | Nte n  -> stack.(sp) <- StackNte(n); exec fp (sp+1) (pc+1)
  | Bln b  -> stack.(sp) <- StackBln(b); exec fp (sp+1) (pc+1)
  | Ivl i  -> stack.(sp) <- StackIvl(i); exec fp (sp+1) (pc+1)
  | Lst (t, l, c) -> stack.(sp) <- StackLst(t, l, c); exec fp (sp+1) (pc+1)
  | Drp    -> exec fp (sp-1) (pc+1)
  | Una op -> let op1 = stack.(sp-1) in stack.(sp-1) <- (match op with
      Not -> (match op1 with
                StackBln(b) -> StackBln(not b)
              | _ -> raise(Failure("invalid unary operand type")))
      | _ -> raise(Failure("invalid unary operator")));
    exec fp (sp) (pc+1)
  | Bin op -> let op1 = stack.(sp-2) and op2 = stack.(sp-1) in     
      stack.(sp-2) <- (match op with
        Add     -> (match (op1, op2) with
            (StackInt(i1), StackInt(i2)) -> StackInt(i1+i2)
          | (StackNte(n1), StackIvl(i1)) ->
            (let int_of_ivl ivl = (match ivl with Whole -> 2 | Half -> 1) in
                  StackNte(n1 + int_of_ivl i1))
          | _ -> raise(Failure("invalid add operands")))
      | Sub -> (match (op1, op2) with
            (StackInt(i1), StackInt(i2)) -> StackInt(i1-i2)
          | (StackNte(n1), StackIvl(i1)) ->
            (let int_of_ivl ivl = (match ivl with Whole -> 2 | Half -> 1) in
                  StackNte(n1 - int_of_ivl i1))
          | _ -> raise(Failure("invalid sub operands")))
      | Mult -> (match (op1, op2) with
            (StackInt(i1), StackInt(i2)) -> StackInt(i1*i2)
        | _ -> raise(Failure("invalid mul operands")))
      | Div -> (match (op1, op2) with
            (StackInt(i1), StackInt(i2)) -> StackInt(i1/i2)
        | _ -> raise(Failure("invalid div operands")))
      | Mod     -> (match (op1, op2) with
          (StackInt(i1), StackInt(i2)) -> StackInt(i1 mod i2)
        | _ -> raise(Failure("invalid mod operands")))
      | Equal   ->  (match (op1, op2) with
          (StackBln(b1), StackBln(b2)) -> StackBln(b1 == b2)
        | (StackInt(i1), StackInt(i2)) -> StackBln(i1 == i2)
        | (StackNte(n1), StackNte(n2)) -> StackBln(n1 == n2)
        | _ -> raise(Failure("stack corrupt")))
      | Neq     ->  (match (op1, op2) with
          (StackBln(b1), StackBln(b2)) -> StackBln(b1 != b2)
        | (StackInt(i1), StackInt(i2)) -> StackBln(i1 != i2)
        | (StackNte(n1), StackNte(n2)) -> StackBln(n1 != n2)
        | _ -> raise(Failure("stack corrupt")))
      | Less    ->  (match (op1, op2) with
          (StackInt(i1), StackInt(i2)) -> StackBln(i1 < i2)
        | (StackNte(n1), StackNte(n2)) -> StackBln(n1 < n2)
        | _ -> raise(Failure("stack corrupt")))
      | Leq     ->  (match (op1, op2) with
          (StackInt(i1), StackInt(i2)) -> StackBln(i1 <= i2)
        | (StackNte(n1), StackNte(n2)) -> StackBln(n1 <= n2)
        | _ -> raise(Failure("stack corrupt")))
      | Greater ->  (match (op1, op2) with
          (StackInt(i1), StackInt(i2)) -> StackBln(i1 > i2)
        | (StackNte(n1), StackNte(n2)) -> StackBln(n1 > n2)
        | _ -> raise(Failure("stack corrupt")))
      | Geq     ->  (match (op1, op2) with
          (StackInt(i1), StackInt(i2)) -> StackBln(i1 >= i2)
        | (StackNte(n1), StackNte(n2)) -> StackBln(n1 >= n2)
        | _ -> raise(Failure("stack corrupt")))
      | Or      -> (match (op1, op2) with
          (StackBln(b1), StackBln(b2)) -> StackBln(b1 || b2)
        | _ -> raise(Failure("stack corrupt")))
      | And     -> (match (op1, op2) with
          (StackBln(b1), StackBln(b2)) -> StackBln(b1 && b2)
        | _ -> raise(Failure("stack corrupt")))
      | Shl     -> (match (op1, op2) with
          (StackNte(n), StackInt(i)) -> StackNte(n - (12*i))
        | _ -> raise(Failure("stack corrupt")))
      | Shr     -> (match (op1, op2) with
          (StackNte(n), StackInt(i)) -> StackNte(n + (12*i))
        | _ -> raise(Failure("stack corrupt")))
      | Deref -> (match (op1, op2) with
          (StackLst(Int, l, c), StackInt(i)) -> StackInt(List.nth c i)
        | (StackLst(Bool, l, c), StackInt(i)) ->
                        (match List.nth c i with
                          0 -> StackBln(false) | _ -> StackBln(true))
        | (StackLst(Note, l, c), StackInt(i)) ->
                        (StackNte(List.nth c i))
        | _ -> print_stack stack sp 10; raise(Failure("stack corrupt")))
      | Concat -> (match (op1, op2) with
          (StackInt(i), StackLst(Int, l, c)) -> StackLst(Int, l+1, i::c)
        | (StackBln(b), StackLst(Bool, l, c)) ->
                        (let i = match b with true -> 1 | false -> 0 in
                          StackLst(Bool, l+1, i::c))
        | (StackNte(n), StackLst(Note, l, c)) ->
                        StackLst(Note, l+1, n::c)
        | _ -> raise(Failure("stack corrupt")))
      | _ -> raise (Failure("unimplemented binop")));
      exec fp (sp-1) (pc+1)
  | Dbg     -> print_stack stack sp 5; exec fp (sp) (pc+1)
  | Nop     -> exec fp sp (pc+1)
  | Lod i   -> stack.(sp)   <- globals.(i)  ; exec fp (sp+1) (pc+1)
  | Str i   -> globals.(i)  <- stack.(sp-1) ; exec fp sp     (pc+1)
  | Lfp i   -> stack.(sp)   <- stack.(fp+i) ; exec fp (sp+1) (pc+1)
  | Sfp i   -> stack.(fp+i) <- stack.(sp-1) ; exec fp sp     (pc+1)
  (* len *)
  | Jsr(-1) -> (let l = stack.(sp-1) in match l with
                    StackLst(_, n, _) -> stack.(sp-1) <- StackInt(n)
                   | _ -> raise(Failure("stack corrupt")));
                   exec fp sp (pc+1)
  (* octave *)
  | Jsr(-2) -> (let n = stack.(sp-1) in match n with
                  StackNte(n) when (n <= 2) -> stack.(sp-1) <-
                          StackInt(0)
                | StackNte(n) when (n > 2) -> stack.(sp-1) <-
                          StackInt(((n-3)/12)+1)
                | _ -> raise (Failure("stack corrupt")));
                exec fp sp (pc+1)
  (* pitch *)
  | Jsr(-3) ->  let rec pitchshift i delta = (match i with
                  i when (i < 39 || i > 50) -> pitchshift (i + delta) delta
                | _ -> i) in
                (let n = stack.(sp-1) in match n with
                  StackNte(n) when (n > 50) ->
                          stack.(sp-1) <- StackNte((pitchshift n (-12)))
                | StackNte(n) when (n < 39) ->
                          stack.(sp-1) <- StackNte((pitchshift n 12))
                | StackNte(_) -> stack.(sp-1) <- n
                | _ -> raise(Failure("stack corrupt")));
                exec fp sp (pc+1)
  (* print *)
  | Jsr(-4) -> (let f = stack.(sp-1) in match f with
                  StackInt(i) -> print_endline (string_of_int i)
                | StackSng(s) -> print_endline s
                | StackNte(n) -> print_endline (notestring_of_int n)
                | StackBln(b) -> print_endline (string_of_bool b)
                | StackLst(Bool, n, c) -> print_endline ("[" ^
                    String.concat ";"
                    (List.map (fun i -> if (i = 0) then "false" else "true") c)
                    ^ "]")
                | StackLst(Int, n, c) -> print_endline ("[" ^
                    String.concat ";"
                    (List.map string_of_int c) ^ "]")
                | StackLst(Note, n, c) -> print_endline ("[" ^
                    String.concat ";"
                    (List.map notestring_of_int c) ^ "]")
                | _ -> raise (Failure("illegal argument to print")));
                exec fp sp (pc+1)
  (* reverse *)
  | Jsr(-5) -> (let l = stack.(sp-1) in match l with
                     StackLst(t, l, c) ->
                           stack.(sp-1) <- StackLst(t, l, List.rev c)
                   | _ -> raise(Failure("stack corrupt")));
                   exec fp sp (pc+1)

  | Jsr i   -> stack.(sp)   <- RetAddr(pc+1)    ; exec fp (sp+1) i
  | Dup     -> stack.(sp) <- stack.(sp-1); exec fp (sp+1) (pc+1)
  | Efr     -> stack.(sp) <- StackSng("stub"); exec fp (sp+1) (pc+1)
  | Swp     -> let t1 = stack.(sp-1) and t2 = stack.(sp-2) in
               stack.(sp-2) <- t1; stack.(sp-1) <- t2; exec fp sp (pc+1)
  | Ent i   -> stack.(sp)   <- StackPos(fp)     ; exec sp (sp+i+1) (pc+1)
  | Rts i   -> let new_fp = stack_pos stack.(fp) and
                   new_pc = ret_addr stack.(fp-1) in
               stack.(fp-i-1) <- stack.(sp-1) ; exec new_fp (fp-i) new_pc
  | Beq i   ->
      exec fp (sp-1) (pc + if stack.(sp-1) = StackBln(true) then i else 1)
  | Bne i   ->
      exec fp (sp-1) (pc + if stack.(sp-1) = StackBln(false) then i else 1)
  | Bra i   -> exec fp sp (pc+i)
  | Hlt     -> ()

  in exec 0 0 0
