%{ open Ast %}

/* assorted punctuation */
%token SEMI COLON LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET COMMA QUOTE
/* intervals and other basic types */
%token UP DOWN WHOLE HALF
/* operators */
%token CONS CONSEQ
%token AND OR NOT MOD PLUS MINUS TIMES DIVIDE DECL ASSIGN PLUSEQ MINUSEQ
%token SHL SHR
/* comparators */
%token EQ NEQ LT LEQ GT GEQ
/* keywords */
%token RETURN IF ELSE FOR IN TO WHILE INT BOOL NOTE LIST VOID

%token <int> LITERAL
%token <bool> BOOLLITERAL
%token <string> NOTELITERAL
%token <string> STRINGLITERAL
%token <string> ID
%token NEW
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN PLUSEQ MINUSEQ CONSEQ
%right CONS
%left AND OR
%left EQ NEQ
%left LT GT LEQ GEQ
%left SHL SHR
%left PLUS MINUS
%left TIMES DIVIDE MOD
%right UMINUS NOT
%left LBRACKET

%start program
%type <Ast.program> program

%%

program:
  /* nothing */ { [], [] }
  | program vdecl { fst $1 @ [$2], snd $1 }
  | program fdecl { fst $1, ($2 :: snd $1) }

typedecl:
  formaltype { $1 }
  | VOID { Void }

formaltype:
    BOOL { Bool }
  | INT { Int }
  | NOTE { Note }
  | LIST LT BOOL GT { List(Bool) }
  | LIST LT INT GT { List(Int) }
  | LIST LT NOTE GT { List(Note) }

formals_opt:
  /* nothing */ { [] }
  | formal_list { List.rev $1 }

formal_list:
    formaltype ID { [($1, $2)] }
  | formal_list COMMA formaltype ID { ($3, $4)::$1 }

actuals_opt:
  /* nothing */   { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }

list_members:
    expr { [$1] }
    | list_members SEMI expr { $3 :: $1 }

fdecl:
  typedecl ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE {
        { rettype = $1;
          fname = $2;
          formals = $4;
          body = List.rev $7;
        } }

stmt_list:
  /* nothing */ { [] }
  | stmt_list stmt { $2 :: $1 }

interval:
    WHOLE { Whole }
  | HALF  { Half }

stepexpr:
    expr { $1 }
  | UP interval { StepExpr(Up, $2) }
  | DOWN interval { StepExpr(Down, $2) }

stmt:
    ID DECL expr SEMI { Declaration($1, $3, false) }
  | expr SEMI { Expr($1) }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7) }
  | WHILE LPAREN expr RPAREN stmt { While ($3, $5) }
  | FOR LPAREN ID DECL expr TO expr RPAREN stmt { For($3, $5, $7, Noexpr, $9) }
  | FOR LPAREN ID DECL expr TO expr COLON stepexpr RPAREN stmt
      { For($3, $5, $7, $9, $11) }
  | FOR LPAREN ID IN expr RPAREN stmt { ForEach ($3, $5, $7) }
  | RETURN expr_opt SEMI  { Return($2) }

vdecl:
  ID DECL expr SEMI {
          { vname = $1;
            value = $3;
          } }

expr_opt:
  /* nothing */ { Noexpr }
  | expr        { $1 }

step:
    HALF  { Half }
  | WHOLE { Whole }

expr:
    LITERAL { Literal($1) }
  | ID      { Id($1) }
  | BOOLLITERAL { BoolLiteral($1) }
  | NOTELITERAL { NoteLiteral(noteint_of_str $1) }
  | STRINGLITERAL  { StringLiteral($1) }
  | LIST LT BOOL GT { EmptyList(Bool) }
  | LIST LT INT GT { EmptyList(Int) }
  | LIST LT NOTE GT { EmptyList(Note) }
  | LBRACKET list_members RBRACKET { ListLiteral(List.rev $2) }
  | MINUS expr %prec UMINUS { Binop(Literal(0), Sub, $2) }
  | NOT expr         { Unaryop(Not, $2) }
  | expr LBRACKET expr RBRACKET { Binop($1, Deref, $3) }
  | expr CONS   expr { Binop ($1, Concat, $3) }
  | expr PLUS   step { Binop($1, Add, IntervalLiteral($3)) }
  | expr MINUS  step { Binop($1, Sub, IntervalLiteral($3)) }
  | expr PLUS   expr { Binop($1, Add,   $3) }
  | expr MINUS  expr { Binop($1, Sub,   $3) }
  | expr TIMES  expr { Binop($1, Mult,  $3) }
  | expr DIVIDE expr { Binop($1, Div,   $3) }
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater,  $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3) }
  | expr MOD    expr { Binop($1, Mod, $3) }
  | expr AND    expr { Binop($1, And, $3) }
  | expr OR     expr { Binop($1, Or, $3) }
  | expr SHL    expr { Binop($1, Shl, $3) }
  | expr SHR    expr { Binop($1, Shr, $3) }
  | ID PLUSEQ   step { Assign($1, Binop(Id($1), Add, IntervalLiteral($3))) }
  | ID MINUSEQ  step { Assign($1, Binop(Id($1), Sub, IntervalLiteral($3))) }
  | ID PLUSEQ   expr   { Assign($1, Binop(Id($1), Add, $3)) }
  | ID MINUSEQ  expr   { Assign($1, Binop(Id($1), Sub, $3)) }
  | ID CONSEQ   expr   { Assign($1, Binop($3, Concat, Id($1))) }
  | ID ASSIGN expr   { Assign($1, $3) }
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN { $2 }
